/*
 * vectors.cpp
 *
 *  Created on: 14-lug-2009
 *      Author: gio
 */

#include <iostream>
#include <JTC/jtc.h>
#include "buffer.h"


using namespace std;

#define PROCESSORS 4


int* sum(int* X, int* Y, int N)
{
	int* result = new int[N];
	for( int i = 0; i < N; i ++)
	{
		result[i] = X[i] + Y[i];
	}
	return result;
}
int* prod(int* X, int* Y, int N)
{
	int* result = new int[N];
	for( int i = 0; i < N; i ++)
	{
		result[i] = X[i] * Y[i];
	}
	return result;
}


class Request
{
public:
	int* X;
	int* Y;
	int N;
	/*
	 *  SommaProd = true -> '+'
	 *  SommaProd = false -> '*'
	 */
	bool sommaProd;

public:
	Request(int* X,int* Y,int N,bool sp)
	{
		this->X = X;
		this->Y = Y;
		this->N = N;
		this->sommaProd = sp;
	}
};

class Barrier : public JTCMonitor

{
	int nthreads;
	int threadsFinished;
public:
	Barrier(int nt)
	{
		nthreads = nt;
		threadsFinished = 0;
	};
	void done()
	{
		JTCSynchronized sync(*this);
		threadsFinished++;
		if (threadsFinished == nthreads)
			notify();
	}
	void waitForCompletion()
	{
		JTCSynchronized sync(*this);
		if (threadsFinished < nthreads)
			wait();
	}
};


class Worker : public JTCThread
{
	Buffer<Request*>* bufRichieste;
	Buffer<int*>* bufRisultati;
    Barrier* b;

public:
	Worker(Buffer<Request*>* bric, Buffer<int*>* bris, Barrier* b)
	{
		bufRichieste = bric;
		bufRisultati = bris;
		this->b = b;
	}
	virtual void run()
	{
		Request* r;
		while(r = bufRichieste->get())
		{
	        //cout << "Computing " << r << endl;
			bufRisultati->put( r->sommaProd ? sum(r->X,r->Y,r->N) : prod(r->X,r->Y,r->N) );
			delete r;
		}
		b->done();
	}

};

int* expr(int** X, int M, int N)

{
    Buffer<Request*>* bufRichieste = new Buffer<Request*>(0);
    Buffer<int*>* bufRisultati = new Buffer<int*>(0);
    Barrier* b = new Barrier(PROCESSORS);
    Worker* t[PROCESSORS];
    for(int i=0; i < PROCESSORS; i++)
    {
        t[i] = new Worker(bufRichieste,bufRisultati,b);
        t[i]->start();
    }
    /*
     * Giro iniziale di M moltiplicazioni
     */
    for(int i=0; i < M; i++)
    {
    	bufRichieste->put(new Request(X[i],X[i],N,false));
    }
    /*
     * Secondo giro di M-1 somme
     */
    for(int i = 0; i < M-1; i++) {
    	int* r0 = bufRisultati->get();
		int* r1 = bufRisultati->get();
		bufRichieste->put(new Request(r0,r1,N,true));
    }
    int* result =  bufRisultati->get();
    /*
     * Segnala ai worker che non c' pi nulla da prendere in bufRichieste.
     */

    //cout << "Stopping Threads..." << endl;
    bufRichieste->close();
    //cout << "Threads notified" << endl;

    /*
     * Attende the tutti i worker abbiano terminato. Garantisce che al momento delle successive delete non
     * ci siano worker che ancora accedono ai buffer.
     */

    b->waitForCompletion();
    //cout << "Threads terminated." << endl;
    delete bufRisultati;
    delete bufRichieste;
    delete b;
    return result;
}

int X[] = {1, 2, -1, 0, 3, 0, 0, 0, 1, 0, 0, 0, 0};
int Y[] = {1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0};
int Z[] = {1, 2, -1, 0, 3, 0, 0, 0, 1, 0, 0, 0, 0};
int K[] = {1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0};

int A[] = {1, 2, -1, 0, 3, 0, 0, 0, 1, 0, 0, 0, 0};
int B[] = {1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0};
int C[] = {1, 2, -1, 0, 3, 0, 0, 0, 1, 0, 0, 0, 0};
int D[] = {1, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1};


int* T[] = { X, Y, Z, K, A, B, C, D };

int main()

{
   JTCInitialize init;
   int* R = expr(T,8,13);
   for(int i = 0; i < 13; i++)
   {
	   cout << R[i] << " ";
   }
   cout << endl;
   return 0;
}

